// PLACE YOUR HEADER BLOCK COMMENT HERE


// LIST class
//
// CONSTRUCTION: with no initializer
// Access is via LIST_ITR class
//
// PUBLIC OPERATIONS:
//
// int size()                   --> Returns number of items in list
// boolean isEmpty()            --> Return true if empty; else false
//
// LIST_ITR first()             --> Return first position
// LIST_ITR last()              --> Return last position
// LIST_ITR find(x)             --> Return position that views
//                                   first occurence of x
// void insertLast(x)           --> Insert x at tail of list
// void insertFirst(x)          --> Insert x at head of list
// void removeLast()            --> Remove element at tail of list
// void removeFirst()           --> Remove element at head of list 
//
// void display(sout)           --> Display list on stream sout
// void debugDisplay(sout)      --> Debug (detailed) display on
//                                   stream sout
// void insertAfter(p, x)       --> Insert x after current iterator
//                                   position p
// void insertBefore(p, x)      --> Insert x before current iterator
//                                   position p
// void remove(x)               --> Remove first occurrence of x
// void remove(p)               --> Remove item at iterator position p
// void makeEmpty()             --> Remove all items
 

#ifndef LIST_H
#define LIST_H


#include <iostream.h>

#include "LIST_NODE.h"
#include "LIST_ITR.h"

template <class T>
class LIST;			// prototype declaration

template <class T>
class LIST_ITR;		// prototype declaration


template <class T>
class LIST
{
	public:
		// constructor and destructor
		LIST();
		~LIST();

		// inspectors and accessors
		int size() const;
		bool isEmpty() const;

		// iterator facilitators
		LIST_ITR<T> first() const;
		LIST_ITR<T> last() const;
		LIST_ITR<T> find(const T &val) const;

		// convenience functions
		void insertLast(const T &val);
		void insertFirst(const T &val);
		void removeLast();
		void removeFirst();

		// (stream) insertion facilitators
		void display(ostream &sout) const;
		void debugDisplay(ostream &sout) const;

		// list facilitators
		void insertBefore(const LIST_ITR<T> &p, const T &val);
		void insertAfter(const LIST_ITR<T> &p, const T &val);
		void remove(const LIST_ITR<T> &p);
		void remove(const T &val);
		void makeEmpty();

		// member assignment and copy construction
		LIST& operator=(const LIST &rhs);
		LIST(const LIST &rhs);

	private:
		// data members
		LIST_NODE<T> *head;
		LIST_NODE<T> *tail;
		int listLength;
};



template <class T>
LIST<T>::LIST()
{
	// Construct an empty list
	head = NULL;
	tail = NULL;
	listLength = 0;
}


template <class T>
LIST<T>::~LIST()
{
	// destructor

	// call makeEmpty() member function to remove all elements from the list
	makeEmpty();
}


template <class T>
int LIST<T>::size() const
{
	// returns the number of elements in the list
	
	// YOUR CODE GOES HERE
}


template <class T>
bool LIST<T>::isEmpty() const
{
	// returns true if list is empty, false otherwise
	return listLength == 0;

	// alternatively:
	// return head == NULL;
}


template <class T>
LIST_ITR<T> LIST<T>::first() const
{
	// returns an iterator pointing to the first element in the list
	// this operation is valid for empty lists

	LIST_ITR<T> result(head);
	return result;
}


template <class T>
LIST_ITR<T> LIST<T>::last() const
{
	// returns an iterator pointing to the last element in the list
	// this operation is valid for empty lists

	// YOUR CODE GOES HERE
}


template <class T>
LIST_ITR<T> LIST<T>::find(const T &val) const
{
	// returns an iterator corresponding to the first node containing item val
	// returns iterator isPastEnd if item is not found

	LIST_NODE<T> *position = head;

	while ((position != NULL) && (position->element != val))
	{
		position = position->next;
	}

	// create an iterator that refers to the position and return it
	LIST_ITR<T> result(position);

	return result;
}


template <class T>
void LIST<T>::insertLast(const T &val)
{
	// inserts an element at the tail of the list

	LIST_NODE<T> *newNode = new LIST_NODE<T>(val);
	
	if (tail == NULL)
	{
		head = newNode;
	}
	else
	{
		tail->next = newNode;
		newNode->prev = tail;
	}

	tail = newNode;
	++listLength;
}


template <class T>
void LIST<T>::insertFirst(const T &val)
{
	// inserts an element at the head of the list 

	// YOUR CODE GOES HERE
	// don't forget to adjust the data member that is tracking the list size
}


template <class T>
void LIST<T>::removeLast()
{
	// removes the element at the tail of the list

	// YOUR CODE GOES HERE
	// don't forget to adjust the data member that is tracking the list size
}


template <class T>
void LIST<T>::removeFirst()
{
	// removes the element at the head of the list

	// YOUR CODE GOES HERE
	// don't forget to adjust the data member that is tracking the list size
}


template <class T>
void LIST<T>::display(ostream &sout) const
{
	// prints out all the elements in the list
	LIST_NODE<T> *ptr = head;

	sout << "[";

	while (ptr != NULL)
	{
		sout << " " << ptr->element;
		ptr = ptr->next;
	}

	sout << " ]" << endl;

}


template <class T>
void LIST<T>::debugDisplay(ostream &sout) const
{
	// for debugging purposes only
	// called like: listA.debug_output(cout);

	// prints out all the elements in the list
	// with detailed information

	// each node in the list appears in the following form
	// address of node: | address of prev | value of element | address of next |

	sout << "Detailed Debugging Display -----------------------------" << endl;
	sout << endl;
	sout << "Value of listLength: " << listLength << endl;
	sout << "Address of head pointer: " << head << endl;
	sout << "Address of tail pointer: " << tail << endl;
	sout << endl;

	sout << "address of node: | address of prev | ";
	sout << "value of element | address of next |" << endl;
	sout << endl;

	LIST_NODE<T> *ptr = head;

	while (ptr != NULL)
	{
		sout << ptr << ": | ";
		sout << ptr->prev << " | " << ptr->element << " | " << ptr->next;
		sout << " |" << endl;

		ptr = ptr->next;
	}

	sout << "--------------------------------------------------------" << endl;
	sout << endl;
}


template <class T>
void LIST<T>::insertBefore(const LIST_ITR<T> &p, const T &val)
{
	// inserts a new element (val) into the list before position p

	// 'position' points to the node we wish to insert before
	LIST_NODE<T> *position = p.current;

	
	// YOUR CODE GOES HERE
	// don't forget to adjust the data member that is tracking the list size
}


template <class T>
void LIST<T>::insertAfter(const LIST_ITR<T> &p, const T &val)
{
	// inserts a new element (val) into the list after position p

	// 'position' points to the node we wish to insert after
	LIST_NODE<T> *position = p.current;

	// YOUR CODE GOES HERE
	// don't forget to adjust the data member that is tracking the list size
}


template <class T>
void LIST<T>::remove(const LIST_ITR<T> &p)
{
	// removes an element at position p

	// 'position' points to the node we wish to remove
	LIST_NODE<T> *position = p.current;


	// YOUR CODE GOES HERE
	// don't forget to adjust the data member that is tracking the list size
}


template <class T>
void LIST<T>::remove(const T &val)
{
	// removes the node containing the first occurrence of val

	// YOUR CODE GOES HERE
	// don't forget to adjust the data member that is tracking the list size
}


template <class T>
void LIST<T>::makeEmpty()
{
	// removes all nodes from the list

	LIST_NODE<T> *newHead;

	while (head != NULL)
	{
		newHead = head->next;
		delete head;
		head = newHead;
	}

	head = NULL;
	tail = NULL;
	listLength = 0;
}


template <class T>
LIST<T>& LIST<T>::operator=(const LIST &rhs)
{
	// memberwise assignment operator
	// this function gets called when you have something in the client code
	// like: listA = listB
	// "this" is listA and rhs is listB

	makeEmpty();
	*this = rhs
}


template <class T>
LIST<T>::LIST(const LIST &rhs)
{
	// copy constructor
	// this function gets called when you have something in the client code
	// like: LIST<int> listA(listB); or LIST<int> listA = listB;
	// "this" is listA and rhs is listB

	if (this != &rhs)
	{
		makeEmpty();

		LIST_ITR<T> itr = rhs.first();

		while (!itr.isPastEnd())
		{
			insertLast(itr.retrieve());
			itr.advance();
		}

	}

}

#endif
